/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller   *
 *   j.keller@imperial.ac.uk   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Symmetries.h"
//TODO : a few things must be placed in the header files instead of the cpp file.


// Main functions
vector<vector<int> > homogenousPoly(int n, int k)
{           // homogenous monomials of deg k in n+1 variables
  int i,j;
  vector<int>  zero(n+1), one;
  vector<vector<int> > f, t;
  if (k<0)  return f;         // takes care of <0
  if (k==0) {f.push_back(zero); return f;}                  // initializes the function
  if (n==0) {one.push_back(k); f.push_back(one); return f;} // initializes the function
  for(i=0; i<=k; i++)
  {
    t=homogenousPoly(n-1,k-i);
    for(j=0; j<t.size(); j++) {t[j].push_back(i); f.push_back(t[j]);}
  }
  return f;
}

vector<vector<int> > trimSections(int n, int k)  //removes the sections that are redundant
{                                   //the rel's of X replace \prod x_i
  vector<vector<int> > f, v = homogenousPoly(n, k);
  for(int j=0; j<v.size(); j++) if(prod(v[j])==0) f.push_back(v[j]);
  return f;
}


template<class T> void reOrder(vector<T> &v, const vector<T> &S_n)
{                             // reorder the sections so that the S_n orbits are next to each other
  set<T> t, r(v.begin(),v.end());   // make a set out of v, T = vector<int>
  typedef typename set<T>::iterator sIter;
  sIter i,j;

  v.clear();                  // reuse v
  while(!r.empty())           // r is what is left
  {
    j=r.begin();            // go to the begining of the set
    t=generateOrbit(*j, S_n);     // generate the orbit of the 1st element under S_n
    for(i=t.begin(); i!=t.end(); i++) {v.push_back(*i); j=r.find(*i); r.erase(j);}
  }                   // remove the oribit of this from the initial set
}

template<class T> int oneAction(const T &a, const T &b, const T &g)      // T=vector<int>
{                             // compute the behaviour of s_a \bar s_b under the action of g\in Z_N
  int sum=0; for(int i=0; i<a.size(); i++) sum += (a[i]-b[i])*g[i]; return sum;
}

template<class T> int checkSymm(const T &a, const T &b, const vector<T> &g, int N)   // T=vector<int>
{                                   // check whether the symm is preserved return 1 if yes
  for(int i=0; i<g.size(); i++) if( oneAction(a,b,g[i])%N != 0) return 0;
  return 1;
}

template<class T> vector<T> AllPairs(T &S, int N)
{           // T=vector<vector<int> >
  typedef vector<int> V;
  T S_n=S_N<vector<int> >(N), gG=makeGensOfG<vector<int> >(N-1);
  int i,j,k,l, NS=S.size(), Tot=0;
  int J[2]; 
  MMatrix<int> G(NS,NS);
  vector<set<V> > pairs;
  vector<T> f;
  set<V> tM;
  V v[2], t;

  reOrder(S, S_n);
  for(i=0; i<NS; i++) for(j=0; j<NS; j++) {G[i][j]=0; Tot +=checkSymm(S[i],S[j],gG,N);}  // Initialize matrix to 0
  for(i=0; i<NS; i++)
    for(j=0; j<NS; j++)
      if(G[i][j]==0)                      // For every possibile combination
        if( checkSymm(S[i],S[j],gG,N) == 1)     // check if (i,j)th component is 0 or not. If it's \neq 0,
  {                               // then ignore it, since it has already been take care of.
    tM.clear();                     // If 0 but could be non-zero, then generate the oribt and
    for(l=0; l<S_n.size(); l++)           // write 1. Moreover, write these pairs into a vector of sets.
    {
      v[0]=permute(S_n[l],S[i]); v[1]=permute(S_n[l],S[j]);
      for(k=0; k<2; k++)  J[k]=find(v[k],S);
      G[J[0]][J[1]]=1;
      t.clear(); for(k=0; k<2; k++) t.push_back(J[k]);
      tM.insert(t);
    }
    pairs.push_back(tM);
  }
  for(i=0; i<pairs.size(); i++) {f.push_back(SetToVector(pairs[i])); Tot -=f[i].size();} // converts the sets into vectors
///  assert(Tot==0);       // check that we got all the pairs
  return f;
}

template<class T> int compare(const T &v1, const T &v2)  // T=vector<vector<int> >
{
  int k;
///  assert(v1.size()==v2.size());
  for(int i=0; i<v1.size(); i++) {k = find(swap(v1[i]), v2); if(k==-1) return 0;}
  return 1;
}

template<class T> int Slim(vector<T>  &v) // T=vector<int>
{                 // throw away the redundant parts of a pairing
  T t;                  // i.e. if (i,j) and (j,i) both appear in the same group, then keep only
  vector<T> f;          // the upper triangular one
  int k, M=0;

  for(int i=0; i<v.size(); i++)
  {
    t=v[i];
    k = find(swap(t), v);
    if(t[0]>t[1] && k!=-1) M +=1; else f.push_back(t);
  }
  ///if(M!=0) assert(v.size()==2*f.size());  // check if the theory is correct
  v=f;
  return M;
}


//TODO this should not give a print out. One should take care of that in in-output.h
vector<vector<int> > getThePairs(vector<vector<int> > &Sec, int N, vector<int> &Block, int NS)
{
  typedef vector<int> V;
  vector<vector<V> > v=AllPairs(Sec,N);
  vector<vector<V> >::iterator i,k;
  int j, ti;
  V t;
  vector<V> w;

  int B=0, B1=0;              // this is just a printout, and count the # of pairs
  for(i=v.begin(); i!=v.end(); i++) {ti=(*i).size(); B += ti; cout<< ti<<":  "; print(*i);}
  cout<< B <<"\n\n";

  i=v.begin();
  while(i!=v.end())
  {
    for(j=0; j<(*i).size(); j++)
    {
      t=(*i)[j];        // get all the vectors then start trimming.
      if(t[0]!=t[1])          // If the pair (t_0,t_1) is different, then see if the reverse appers
      {               // in another group. If it does then do more work.
        t = swap(t);          // First swap the 2 vector (t_0,t_1)
        k=i+1;
        while( k!=v.end() )
        {
          ti=find(t,*k);            // Try to find (t_0,t_1) in a new group.
          if(ti!=-1)                // If found, then do some tests and erase that group from v.
          {
            ///assert(compare(*i,*k)==1);
             v.erase(k);  // But remember how long it was in B.
            ///assert(Slim(*i)==0); 
             v.push_back(*i); v.erase(i); i--;}
          if(k!=v.end()) k++;       // Reorder v such that components w/ friends thrown away are at the end
        }                   // Also test the what is trown is the mirror of what is kept
      }               // the if(k!=v.end()) is needed since we erased
    }
    if(i!=v.end()) i++;
  }
   // Reorder v such that Slim-ed things are at the end, elongate B as well
  for(i=v.begin(); i!=v.end(); i++) if(Slim(*i)!=0) {v.push_back(*i); v.erase(i); i--;}

  j=0;                        // Make a vector of the lengths of each block, and
  for(i=v.begin(); i!=v.end(); i++) // merge the 3 vector into a 2 vector of pairs
  {
    ti=(*i).size(); cout<< ti<<":  "; print(*i); Block.push_back(ti);
    for(j=0; j<ti; j++) w.push_back( (*i)[j] );
    B1 += ti;
  }
  cout<< "\n total ="<< B1 <<"\n\n";
  ///assert(2*(w.size()-NS)==B-NS);
  return w;
}

vector<int> theNumbers(int k, int B)
{ 
  vector<int> L;
  int i;
  while(L.size() < k)
  {
    i = rand()%B;
    if(find(i,L)==-1) L.push_back(i);
  }
  return L;
}

int newEndOfDiag(int NS, VVector<int> Bl, VVector<int> BlO)
{for(int i=0; i<Bl.size(); i++) if(Bl[i]==NS) return Bl[i];}

vector<vector<int> > looseSome(int &ns, vector<vector<int> > L, int k, VVector<int> Bs, VVector<int> Bl, VVector<int> &BsO, VVector<int> &BlO)
{
  vector<vector<int> > v;
  int j;
  vector<int> m;

  for(int i=0; i<Bs.size(); i++)
  {
    BlO[i] = v.size();
    if(Bs[i]>k)
    {
      m = theNumbers(k, Bs[i]);
      for(j=0; j<k; j++) v.push_back( L[ Bl[i]+m[j] ] );
      BsO[i] = k;
    }
    else
    {
      for(j=0; j<Bs[i]; j++) v.push_back( L[Bl[i]+j] );
      BsO[i] = Bs[i];
    }
  }
  ns = newEndOfDiag(ns, Bl, BlO);
  return v;
}


template<class T> vector<T> makeGensOfG(int n)
{           // make the generators of Z_N^D, T = vector<int>
  T t0(n+1,0), t;
  vector<T> f;
  for(int i=0; i<n; i++) {t=t0; t[i]=1; t[n]=-1; f.push_back(t);}
  return f;
}

